import 'dart:async';
import 'dart:convert';
import 'dart:developer';
import 'dart:io';
import 'package:busha_commerce_flutter/src/core/api/api_exception.dart';
import 'package:busha_commerce_flutter/src/core/api/api_routes.dart';
import 'package:busha_commerce_flutter/src/core/services/http_interceptor.dart';
import 'package:busha_commerce_flutter/src/utils/utils.dart';
import 'package:flutter/foundation.dart';
import 'package:http/http.dart' as http;
import 'package:http/io_client.dart';
import 'package:http_interceptor/http/intercepted_client.dart';

class HttpService {

  static List<String> allHosts = [
    BaseUrl.auth.split('//').last,
  ];

  http.Client client = InterceptedClient.build(
    client: IOClient(HttpClient()
      ..badCertificateCallback = ((cert, host, port) {
        final isValidHost = allHosts.contains(host);
        return isValidHost;
      })),
    requestTimeout: const Duration(seconds: 20),
    interceptors: [
      HttpInterceptor(),
    ],
  );

  Future<dynamic> get(
    String route, {
    Map<String, String>? headers,
    bool returnTypeIsFile = false,
  }) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);

        final response = await client.get(url, headers: headers);
        return returnTypeIsFile
            ? _processFileResponse(response)
            : _processResponse(response);
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('Your device is not connected to the internet');
    }
  }

  Future<int> specialGet(
    String route, {
    Map<String, String>? headers,
    bool returnTypeIsFile = false,
  }) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);

        final response = await client.get(url, headers: headers);
        return response.statusCode;
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('Your device is not connected to the internet');
    }
  }

  Future<dynamic> post(String route, dynamic body, dynamic headers) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);

        if (kDebugMode) {
          print(url);
          log(ModelFormatter.formatObject(body).toString());
        }

        final response = await client.post(
          url,
          body: jsonEncode(ModelFormatter.formatObject(body)),
          headers: headers,
        );

        return _processResponse(response);
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('No internet connection');
    }
  }

  Future<dynamic> put(String route, dynamic body, dynamic headers) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);

        final response = await client.put(
          url,
          body: jsonEncode(body),
          headers: headers,
        );
        return _processResponse(response);
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('No internet connection');
    }
  }

  Future<dynamic> patch(String route, dynamic body, dynamic headers) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);
        final response = await client.patch(
          url,
          body: jsonEncode(body),
          headers: headers,
        );
        return _processResponse(response);
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('No internet connection');
    }
  }

  Future delete(String route) async {
    bool isOnline = await hasNetwork();
    if (isOnline) {
      try {
        final url = Uri.parse(route);
        final response = await client.delete(url);
        return _processResponse(response);
      } on SocketException {
        throw FetchDataException('No internet connection 😑');
      } on FormatException {
        throw FetchDataException('Bad response format 👎');
      } on TimeoutException {
        throw FetchDataException('Timeout exceeded');
      }
    } else {
      throw FetchDataException('No internet connection');
    }
  }

  dynamic _processResponse(http.Response response) {
    final jsonBody = response.body.isEmpty ? {} : jsonDecode(response.body);
    final isContentTypeApplicationJSON =
        response.headers['content-type'] != null
            ? response.headers['content-type']!.contains('application/json')
            : false;

    switch (response.statusCode) {
      case 200:
      case 201:
      case 204:
        return jsonBody;

      case 206:
        throw ApiException(jsonBody['message'].toString());
      case 400:
      case 401:
      case 403:
      case 404:
        jsonBody['message'] != null
            ? throw ApiException(jsonBody['message'].toString(),
                meta: jsonBody['meta'])
            : throw ApiException(
                jsonBody['error']['message'].toString(),
                meta: jsonBody['error']['meta'],
              );

      case 500:
      default:
        isContentTypeApplicationJSON
            ? jsonBody['message'] != null
                ? throw ApiException(
                    jsonBody['message'].toString(),
                  )
                : throw ApiException(
                    jsonBody['error']['message'].toString(),
                  )
            : throw FetchDataException(
                'Error occurred while Communicating with Server');
    }
  }

  dynamic _processFileResponse(http.Response response) {
    switch (response.statusCode) {
      case 200:
      case 201:
      case 204:
        return response;

      case 206:
        final jsonBody = response.body.isEmpty ? {} : jsonDecode(response.body);
        throw ApiException(jsonBody['message'].toString());
      case 400:
        final jsonBody = response.body.isEmpty ? {} : jsonDecode(response.body);
        throw ApiException(jsonBody['message'].toString());
      case 401:
      case 403:
        final jsonBody = response.body.isEmpty ? {} : jsonDecode(response.body);
        throw ApiException(jsonBody['message'].toString());
      case 404:
        final jsonBody = response.body.isEmpty ? {} : jsonDecode(response.body);
        throw ApiException(jsonBody['message'].toString());
      case 500:
      default:
        throw FetchDataException(
            'Error occurred while Communicating with Server with StatusCode: ${response.statusCode}');
    }
  }
}

class HttpServiceOverrides extends HttpOverrides {
  /// Valid host platform
  static List<String> allHosts = [
    BaseUrl.auth.split('//').last,
  ];

  /// Override the default certificate even before the app start
  @override
  HttpClient createHttpClient(SecurityContext? context) {
    return super.createHttpClient(context)
      ..badCertificateCallback = ((cert, host, port) {
        final isValidHost = allHosts.contains(host);
        return isValidHost;
      });
  }
}
